import copy
import typing
import functools
import pettingzoo
import numpy as np
import gymnasium as gym
import safe_autonomy_simulation
import safe_autonomy_sims.pettingzoo.docking.reward as r
import safe_autonomy_sims.pettingzoo.docking.utils as utils


class MultiDockingEnv(pettingzoo.ParallelEnv):
    r"""
    ## Description

    This environment is a spacecraft docking problem.
    The goal is for a single deputy spacecraft to navigate
    towards and dock onto a chief spacecraft.

    ## Action Space

    Actions are thruster force values for each of the 3 bi-directional thrusters
    on the x-, y-, and z-axis of the deputy spacecraft's body frame.

    The action space is a `Box(-1, 1, shape=(3,))` where each value in the array
    represents the force applied to the x-, y-, and z-axis of the deputy spacecraft.

    | Index | Action        | Control Min | Control Max | Type (units) |
    |-------|---------------|-------------|-------------|--------------|
    | 0     | x-axis thrust | -1          | 1           | Force (N)    |
    | 1     | y-axis thrust | -1          | 1           | Force (N)    |
    | 2     | z-axis thrust | -1          | 1           | Force (N)    |

    ### Observation Space

    At each timestep, the agent receives the observation,
    $o = [x, y, z, v_x, v_y, v_z, s, v_{limit}]$, where:

    * $x, y,$ and $z$ represent the deputy's position in the Hill's frame,
    * $v_x, v_y,$ and $v_z$ represent the deputy's directional velocity in the Hill's frame,
    * $s$ is the speed of the deputy,
    * $v_{limit}$ is the safe velocity limit given by: $v_{max} + an(d_{chief} - r_{docking})$
        * $v_{max}$ is the maximum allowable velocity of the deputy within the docking region
        * $a$ is the slope of the linear velocity limit as a function of distance from the docking region
        * $n$ is the mean motion constant
        * $d_{chief}$ is the deputy's distance from the chief
        * $r_{docking}$ is the radius of the docking region

    The observation space is a `Box([-np.inf, -np.inf, -np.inf, -np.inf, -np.inf, -np.inf, 0, 0], np.inf, shape=(8,))`.

    | Index | Observation                              | Min     | Max    | Type (units)   |
    |-------|------------------------------------------|---------|--------|----------------|
    | 0     | x position of the deputy in Hill's frame | -np.inf | np.inf | position (m)   |
    | 1     | y position of the deputy in Hill's frame | -np.inf | np.inf | position (m)   |
    | 2     | z position of the deputy in Hill's frame | -np.inf | np.inf | position (m)   |
    | 3     | x velocity of the deputy                 | -np.inf | np.inf | velocity (m/s) |
    | 4     | y velocity of the deputy                 | -np.inf | np.inf | velocity (m/s) |
    | 5     | z velocity of the deputy                 | -np.inf | np.inf | velocity (m/s) |
    | 6     | speed of the deputy                      | -np.inf | np.inf | speed (m/s)    |
    | 7     | safe velocity limit                      | -np.inf | np.inf | velocity (m/s) |


    ## State Transition Dynamics

    The relative motion between the deputy and chief is represented by
    linearized Clohessy-Wiltshire equations [[1]](#1), given by

    $$
        \dot{\boldsymbol{x}} = A {\boldsymbol{x}} + B\boldsymbol{u},
    $$

    where:
    * state $\boldsymbol{x}=[x,y,z,\dot{x},\dot{y},\dot{z}]^T \in \mathcal{X}=\mathbb{R}^6$
    * control $\boldsymbol{u}= [F_x,F_y,F_z]^T \in \mathcal{U} = [-1N, 1N]^3$

    * $$
        A =
    \begin{bmatrix}
    0 & 0 & 0 & 1 & 0 & 0 \\
    0 & 0 & 0 & 0 & 1 & 0 \\
    0 & 0 & 0 & 0 & 0 & 1 \\
    3n^2 & 0 & 0 & 0 & 2n & 0 \\
    0 & 0 & 0 & -2n & 0 & 0 \\
    0 & 0 & -n^2 & 0 & 0 & 0 \\
    \end{bmatrix},
        B =
    \begin{bmatrix}
    0 & 0 & 0 \\
    0 & 0 & 0 \\
    0 & 0 & 0 \\
    \frac{1}{m} & 0 & 0 \\
    0 & \frac{1}{m} & 0 \\
    0 & 0 & \frac{1}{m} \\
    \end{bmatrix},
    $$

    * mean motion constant $n = 0.001027 rad/s$

    ## Rewards

    The reward is the sum of the following terms:

    * $r_t += c(e^{-ad_t} - e^{-ad_{t-1}})$
        * this is a dense reward for approaching the chief
        * $c$ is a scale factor
        * $a$ is the exponential coefficent (can be calculated from a pivot value)
        * $d_t$ is the distance from the chief at time $t$
    * $r_t += -((\delta{v} / m) + b)$
        * this is a dense reward for minimizing fuel usage
        * $\delta{v}$ is the change in velocity
        * $m$ is the mass of the deputy
        * $b$ is a tunable bia term
    * $r_t += v - v_{limit}$ if $v_{limit} < v$, else 0
        * this is a dense reward that punishes velocity constraint violations
        * $v_{limit}$ is the safe velocity limit given by: $v_{max} + an(d_{chief} - r_{docking})$, described above
    * $r += 1 + (1 - (t/t_{max}))$ if docking is successful, else 0
        * this is a sparse reward for minimizing the time required to dock onto the chief
        * $t$ is the current time
        * $t_{max}$ is the maximum episode length before timeout
    * $r += -1.0$ if the agent times out, crashes, or goes out of bounds, else 0
        * this is a sparse reward that punishes the agent for failing to dock onto the chief

    ## Starting State

    At the start of any episode, the state is randomly initialized with the following conditions:

    * chief $(x,y,z)$ = $(0, 0, 0)$
    * docking radius = $0.5 m$
    * deputy position $(x, y, z)$ is converted after randomly selecting the position in polar notation $(r, \phi, \psi)$ using a uniform distribution with
        * $r \in [100, 150] m$
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
        * $x = r \cos{\psi} \cos{\phi}$
        * $y = r \sin{\psi} \cos{\phi}$
        * $z = r \sin{\phi}$
    * deputy $(v_x, v_y, v_z)$ is converted after randomly selecting the velocity in polar notation $(v, \phi, \psi)$ using a Gaussian distribution with
        * $v \in [0, 0.8]$ m/s
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
        * $v_x = v \cos{\psi} \cos{\phi}$
        * $v_y = v \sin{\psi} \cos{\phi}$
        * $v_z = v \sin{\phi}$

    ## Episode End

    An episode will end if any of the following conditions are met:

    * Termination: the agent exceeds a `max_distance = 10000` meter radius away from the chief,
    * Termination: the agent violates the velocity constraint within the docking region (crash),
    * Termination: the velocity limit penalty exceeds -5
    * Truncation: the maximum number of timesteps, `max_timesteps = 2000`, is reached

    The episode is considered done and successful if and only if the agent maneuvers the deputy
    within the docking region while maintaining a safe velocity.

    ## Information

    `step()` and `reset()` return a dict with the following keys:
    * reward_components - a dict of reward component string names to their last computed float values.
    * status - a string descirbing the status of the current episode.

    Statuses for the episode are either "Running" or describe a unique terminal state. Terminal
    states can be one of the following: "Success", "Crash", "Out of Bounds", "Timeout", 
    "Max Velocity Violation".

    ## References

    <a id="1">[1]</a>
    Clohessy, W., and Wiltshire, R., “Terminal Guidance System for Satellite Rendezvous,”
    *Journal of the Aerospace Sciences*, Vol. 27, No. 9, 1960, pp. 653–658.
    """

    def __init__(
        self,
        num_agents: int = 2,
        docking_radius: float = 0.2,
        max_time: int = 2000,
        max_distance: float = 10000,
        max_v_violation: int = 5,
    ) -> None:
        self.possible_agents = [f"deputy_{i}" for i in range(num_agents)]

        # Environment parameters
        self.docking_radius = docking_radius
        self.max_time = max_time
        self.max_distance = max_distance
        self.max_v_violation = max_v_violation

        # Episode level information
        self.prev_state = None
        self.episode_v_violations = 0
        self.reward_components = {a: {} for a in self.possible_agents}
        self.status = {a: "Running" for a in self.possible_agents}

    def reset(
        self, seed: int | None = None, options: dict[str, typing.Any] | None = None
    ) -> tuple[typing.Any, dict[str, typing.Any]]:
        self.rng = np.random.default_rng(seed)
        self.agents = copy.copy(self.possible_agents)
        self._init_sim()  # sim is light enough we just reconstruct it
        self.simulator.reset()
        observations = {a: self._get_obs(agent=a) for a in self.agents}
        infos = {a: self._get_info(agent=a) for a in self.agents}
        self.prev_state = None
        self.episode_v_violations = 0
        return observations, infos

    def step(
        self, actions: dict[str, typing.Any]
    ) -> tuple[typing.Any, typing.SupportsFloat, bool, bool, dict[str, typing.Any]]:
        # Store previous simulator state
        self.prev_state = self.sim_state.copy()

        # Update simulator state
        for a, action in actions.items():
            self.deputies[a].add_control(action)
        self.simulator.step()

        # Get info from simulator
        observations = {a: self._get_obs(agent=a) for a in self.agents}
        rewards = {a: self._get_reward(agent=a) for a in self.agents}
        terminations = {a: self._get_terminated(agent=a) for a in self.agents}
        truncations = {
            a: False for a in self.agents
        }  # used to signal episode ended unexpectedly
        # End episode if any agent is terminated or truncated
        if any(terminations.values() or any(truncations.values())):
            truncations = {a: True for a in self.agents}
            terminations = {a: True for a in self.agents}
            infos = {a: self._get_info(agent=a) for a in self.agents}
            self.agents = []
        else:
            infos = {a: self._get_info(agent=a) for a in self.agents}

        return observations, rewards, terminations, truncations, infos

    def _init_sim(self) -> None:
        # Initialize simulator with chief and deputy spacecraft
        self.chief = safe_autonomy_simulation.sims.spacecraft.CWHSpacecraft(
            name="chief"
        )
        self.deputies = {
            a: safe_autonomy_simulation.sims.spacecraft.CWHSpacecraft(
                name=a,
                position=utils.polar_to_cartesian(
                    r=self.rng.uniform(100, 150),
                    phi=self.rng.uniform(-np.pi / 2, np.pi / 2),
                    theta=self.rng.uniform(0, 2 * np.pi),
                ),
                velocity=utils.polar_to_cartesian(
                    r=self.rng.uniform(0, 0.8),
                    phi=self.rng.uniform(-np.pi / 2, np.pi / 2),
                    theta=self.rng.uniform(0, 2 * np.pi),
                ),
            )
            for a in self.agents
        }
        self.simulator = safe_autonomy_simulation.Simulator(
            frame_rate=1, entities=[self.chief] + list(self.deputies.values())
        )

    def _get_obs(self, agent: str) -> typing.Any:
        deputy = self.deputies[agent]
        v_lim = utils.v_limit(chief_pos=self.chief.position, deputy_pos=deputy.position)
        s = np.linalg.norm(deputy.velocity)
        obs = self.observation_space(agent=agent).sample()
        obs[:3] = deputy.position
        obs[3:6] = deputy.velocity
        obs[6] = s
        obs[7] = v_lim
        return obs

    def _get_info(self, agent: str) -> dict[str, typing.Any]:
        return {
            "reward_components": self.reward_components[agent],
            "status": self.status[agent]
        }

    def _get_reward(self, agent: str) -> float:
        reward = 0
        deputy = self.deputies[agent]

        # Dense rewards
        dist_pivot_reward = r.distance_pivot_reward(
            rel_dist=utils.rel_dist(pos1=self.chief.position, pos2=deputy.position),
            rel_dist_prev=utils.rel_dist(
                pos1=self.prev_state["chief"][0:3], pos2=self.prev_state[agent][0:3]
            ),
        )
        self.reward_components[agent]["distance_pivot"] = dist_pivot_reward
        reward += dist_pivot_reward

        delta_v_reward = r.delta_v_reward(
            v=deputy.velocity, prev_v=self.prev_state[agent][3:6]
        )
        self.reward_components[agent]["delta_v"] = delta_v_reward
        reward += delta_v_reward

        vel_const_reward = r.velocity_constraint_reward(
            v1=deputy.velocity,
            v2=self.chief.velocity,
            v_limit=utils.v_limit(
                chief_pos=self.chief.position, deputy_pos=deputy.position
            ),
        )
        self.reward_components[agent]["velocity_constraint"] = vel_const_reward
        reward += vel_const_reward

        # Sparse rewards
        success_reward = r.docking_success_reward(
            chief=self.chief,
            deputy=deputy,
            t=self.simulator.sim_time,
            vel_limit=utils.v_limit(
                chief_pos=self.chief.position, deputy_pos=deputy.position
            ),
            docking_radius=self.docking_radius,
            max_time=self.max_time,
        )
        self.reward_components[agent]["success"] = success_reward
        reward += success_reward

        timeout_reward = r.timeout_reward(t=self.simulator.sim_time, max_time=self.max_time)
        self.reward_components[agent]["timeout"] = timeout_reward
        reward += timeout_reward

        crash_reward = r.crash_reward(
            chief=self.chief,
            deputy=deputy,
            vel_limit=utils.v_limit(
                chief_pos=self.chief.position, deputy_pos=deputy.position
            ),
            docking_radius=self.docking_radius,
        )
        self.reward_components[agent]["crash"] = crash_reward
        reward += crash_reward

        oob_reward = r.out_of_bounds_reward(
            chief_pos=self.chief.position,
            deputy_pos=deputy.position,
            max_distance=self.max_distance,
        )
        self.reward_components[agent]["out_of_bounds"] = oob_reward
        reward += oob_reward

        return reward

    def _get_terminated(self, agent: str) -> bool:
        deputy = self.deputies[agent]
        # Get state info
        d = utils.rel_dist(pos1=self.chief.position, pos2=deputy.position)
        v = utils.rel_vel(vel1=self.chief.velocity, vel2=deputy.velocity)
        vel_limit = utils.v_limit(
            chief_pos=self.chief.position, deputy_pos=deputy.position
        )
        in_docking = d < self.docking_radius
        safe_v = v < vel_limit
        self.episode_v_violations += -r.velocity_constraint_reward(
            v1=deputy.velocity, v2=self.chief.velocity, v_limit=vel_limit
        )

        # Determine if in terminal state
        oob = d > self.max_distance
        crash = in_docking and not safe_v
        max_v_violation = self.episode_v_violations > self.max_v_violation
        timeout = self.simulator.sim_time > self.max_time
        docked = in_docking and safe_v

        # Update Status
        if crash:
            self.status[agent] = "Crash"
        elif oob:
            self.status[agent] = "Out of Bounds"
        elif timeout:
            self.status[agent] = "Timeout"
        elif max_v_violation:
            self.status[agent] = "Max Velocity Violation"
        elif docked:
            self.status[agent] = "Success"

        return oob or crash or max_v_violation or timeout or docked

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent) -> gym.Space:
        return gym.spaces.Box(
            np.array([-np.inf, -np.inf, -np.inf, -np.inf, -np.inf, -np.inf, 0, 0]),
            np.inf,
            shape=(8,),
        )

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent) -> gym.Space:
        return gym.spaces.Box(-1, 1, shape=(3,))

    @property
    def sim_state(self) -> dict:
        state = {"chief": self.chief.state}
        for a in self.agents:
            state[a] = self.deputies[a].state
        return state



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module defines the measurement and control properties for CWH spacecraft sensors and controllers.
"""
import typing

import gymnasium
import numpy as np
from corl.libraries.property import BoxProp, NestedQuantity, Prop, Quantity
from pydantic import Field, StrictFloat
from typing_extensions import Annotated


class ThrustProp(BoxProp):
    """
    Thrust control properties.

    Attributes
    ----------
    name : str
        Control property name.
    low : list[float]
        Minimum bounds of control input.
    high : list[float]
        Maximum bounds of control input.
    unit : str
        Unit of measurement for control input.
    description : str
        Description of control properties.
    """

    name: str = "thrust"
    low: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [-1.0]
    high: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [1.0]
    unit: str = "newtons"
    description: str = "Direct Thrust Control"


class MomentProp(BoxProp):
    """
    Moment control properties.

    Attributes
    ----------
    name : str
        Control property name.
    low : list[float]
        Minimum bounds of control input.
    high : list[float]
        Maximum bounds of control input.
    unit : str
        Unit of measurement for control input.
    description : str
        Description of control properties.
    """

    name: str = "moment"
    low: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [-0.001]
    high: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [0.001]
    unit: str = "dimensionless"
    description: str = "Direct Moment Control"


class PositionProp(BoxProp):
    """
    Position sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "position"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-10000.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [10000.0] * 3
    unit: str = "m"
    description: str = "Position Sensor Properties"


class RelativePositionProp(BoxProp):
    """
    Relative position sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "relative_position"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-20000.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [20000.0] * 3
    unit: str = "m"
    description: str = "Relative Position Sensor Properties"


class VelocityProp(BoxProp):
    """
    Velocity sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "velocity"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-1000.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [1000.0] * 3
    unit: str = "m/s"
    description: str = "Velocity Sensor Properties"


class RelativeVelocityProp(BoxProp):
    """
    Relative velocity sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "relative_velocity"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-2000.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [2000.0] * 3
    unit: str = "m/s"
    description: str = "Relative Velocity Sensor Properties"


class InspectedPointProp(BoxProp):
    """
    Inspected points sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "inspected_points"
    low: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [0.]
    high: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [100.]
    unit: str = "dimensionless"
    description: str = "Inspected Points Sensor Properties"


class SunAngleProp(BoxProp):
    """
    Sun angle sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "SunAngle"
    low: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [-10000.]
    high: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [10000.]
    unit: str = "rad"
    description: str = "Sun Angle Sensor Properties"


class SunVectorProp(BoxProp):
    """
    Sun vector sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "SunVector"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-1.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [1.0] * 3
    unit: str = "m"
    description: str = "Sun Unit Vector Properties"


class UninspectedPointProp(BoxProp):
    """
    Uninspected points cluster location properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "uninspected_points"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-1.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [1.0] * 3
    unit: str = "m"
    description: str = "Uninspected points cluster location"


class BoolArrayProp(BoxProp):
    """
    bool array properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "bool_array"
    num_points: int = 100
    low: Annotated[typing.List[StrictFloat], Field(min_items=num_points, max_items=num_points)] = [0.0] * num_points
    high: Annotated[typing.List[StrictFloat], Field(min_items=num_points, max_items=num_points)] = [1.0] * num_points
    unit: str = "dimensionless"
    description: str = "Boolean array"


class QuaternionProp(BoxProp):
    """
    Quaternion sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "quaternion"
    low: Annotated[typing.List[StrictFloat], Field(min_items=4, max_items=4)] = [-1.0] * 4
    high: Annotated[typing.List[StrictFloat], Field(min_items=4, max_items=4)] = [1.0] * 4
    unit: str = "dimensionless"
    description: str = "Quaternion Sensor Properties"


class OrientationVectorProp(BoxProp):
    """
    Orientation Unit Vector sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "orientation_unit_vector"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-1.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [1.0] * 3
    unit: str = "dimensionless"
    description: str = "Orientation Unit Vector Sensor Properties"


class RotatedAxesProp(BoxProp):
    """
    Coordinate axis unit vectors rotated into agent frame

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "rotated_axes_unit_vectors"
    low: Annotated[typing.List[StrictFloat], Field(min_items=6, max_items=6)] = [-1.0] * 6
    high: Annotated[typing.List[StrictFloat], Field(min_items=6, max_items=6)] = [1.0] * 6
    unit: str = "dimensionless"
    description: str = "Rotated Axes Unit Vectors Sensor Properties"


class AngularVelocityProp(BoxProp):
    """
    Angular Velocity sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "angular velocity"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-10.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [10.0] * 3
    unit: str = "rad/s"
    description: str = "Angular Velocity Sensor Properties"


class PriorityVectorProp(BoxProp):
    """
    Priority Vector sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "priority vector"
    low: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [-1.0] * 3
    high: Annotated[typing.List[StrictFloat], Field(min_items=3, max_items=3)] = [1.0] * 3
    unit: str = "m"
    description: str = "Priority Vector Sensor Properties"


class PointsScoreProp(BoxProp):
    """
    Inspected points score sensor properties.

    Attributes
    ----------
    name : str
        Sensor property name.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "inspected points score"
    low: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [0.]
    high: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [1.]
    unit: str = "dimensionless"
    description: str = "Inspected Points Score Sensor Properties"


class TupleProp(Prop):
    """
    Tuple space properties

    spaces : tuple
        set of subspaces to combine into a single Tuple space
    """
    spaces: typing.Tuple

    def create_space(self, seed=None) -> gymnasium.spaces.Space:
        """
        Creates MultiDiscrete gymnasium space
        """
        subspaces = tuple(sub_prop.create_space(seed=seed) for sub_prop in self.spaces)
        return gymnasium.spaces.Tuple(subspaces)

    def get_units(self):
        return tuple({prop.name: prop.get_units()} for prop in self.spaces)

    def create_unit_converted_prop(self, unit: str):
        raise RuntimeError(
            f"Prop was told to try and convert to unit {unit} but the only but a Prop is not "
            f"capable of having a unit as it is not a leaf node.  this error should never happen and is just a sanity check"
            "to show this code path is not supported"
        )

    def zero_mean(self) -> Prop:
        raise NotImplementedError

    def scale(self, scale) -> Prop:
        raise NotImplementedError

    def create_quantity(self, value: dict | (float | (int | (list | np.ndarray)))) -> Quantity | NestedQuantity:
        """
        This function taskes in values and will attempt to create either a Quantity or NestedQuantity
        from it, properly applying units along the way
        """
        raise NotImplementedError

    def create_zero_sample(self) -> Quantity | NestedQuantity:
        """
        This function will attempt to return 0 for each leaf node for all properties
        in a tree, as this is usually a safe default.  however if 0 is not in the low
        or high for this space it will return the lowest value possible
        Discrete values will always be the lowest, which should be 0
        """
        raise NotImplementedError

    def create_low_sample(self) -> Quantity | NestedQuantity:
        """
        This function will attempt to return the lowest possible value for each leaf node
        for all properties in a tree.
        """
        raise NotImplementedError


class OrbitStabilityProp(BoxProp):
    """
    Property to hold the orbit stability quantity 2*n*x + v_y
    Highs and lows are obtained manually from this equation and the highs/lows
    of the position and velocity properties and rounded

    Attributes
    ----------
    name : str
        Sensor property name.
    low : list[float]
        Minimum bounds of sensor output.
    high : list[float]
        Maximum bounds of sensor output.
    unit : str
        Unit of measurement for sensor output.
    description : str
        Description of sensor properties.
    """

    name: str = "orbit stability"
    low: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [-1025.0]
    high: Annotated[typing.List[StrictFloat], Field(min_items=1, max_items=1)] = [1025.0]
    unit: str = "m/s"
    description: str = "Orbit Stability Sensor Property"



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning Core (CoRL) Safe Autonomy Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements functions which define the terminal conditions
for the CWH spacecraft environments.
"""
import typing

import gymnasium
import numpy as np
from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator, DoneStatusCodes
from corl.simulators.common_platform_utils import get_platform_by_name
from pydantic.types import PyObject

from safe_autonomy_sims.utils import VelocityConstraintValidator, get_relative_position, get_relative_velocity, max_vel_violation


class MaxDistanceDoneValidator(DoneFuncBaseValidator):
    """
    A configuration validator for the MaxDistanceDoneFunction

    Attributes
    ----------
    max_distance: float
        The maximum tolerated relative distance between deputy and origin before episode termination.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    """

    max_distance: float
    reference_position_sensor_name: str = "reference_position"


class MaxDistanceDoneFunction(DoneFuncBase):
    """
    A done function that determines if the agent is
    beyond a maximum distance from the origin.

    Attributes
    ----------
    config: MaxDistanceDoneValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: MaxDistanceDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        MaxDistanceDoneValidator
            Config validator for the MaxDistanceDoneFunction.
        """
        return MaxDistanceDoneValidator

    def __call__(
        self,
        observation,
        action,
        next_observation,
        next_state,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        # compute distance to reference entity
        platform = get_platform_by_name(next_state, self.config.platform_name)
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        distance = np.linalg.norm(relative_position)

        done = bool(distance > self.config.max_distance)

        if done:
            next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE

        return done


class CrashDoneValidator(DoneFuncBaseValidator):
    """
    A configuration validator for the CrashDoneFunction.

    Attributes
    ----------
    crash_region_radius : float
        The radius of the crashing region in meters.
    velocity_constraint : VelocityConstraintValidator
        Velocity constraint parameters.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    reference_velocity_sensor_name: str
        The name of the sensor responsible for returning the relative velocity of a reference entity.
    """
    crash_region_radius: float
    velocity_constraint: typing.Union[VelocityConstraintValidator, None] = None
    reference_position_sensor_name: str = "reference_position"
    reference_velocity_sensor_name: str = "reference_velocity"


class CrashDoneFunction(DoneFuncBase):
    """
    A done function which determines if deputy has crashed
    with the chief situated at the origin.

    Attributes
    ----------
    config: CrashDoneValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: CrashDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        CrashDockingDoneValidator
            Config validator for the CrashDockingDoneFunction.

        """
        return CrashDoneValidator

    def __call__(
        self,
        observation,
        action,
        next_observation,
        next_state,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        # Get relatative position + velocity between platform and docking region
        platform = get_platform_by_name(next_state, self.config.platform_name)
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        distance = np.linalg.norm(relative_position)

        relative_velocity = get_relative_velocity(platform, self.config.reference_velocity_sensor_name)

        in_crash_region = bool(distance <= self.config.crash_region_radius)

        done = in_crash_region

        if self.config.velocity_constraint is not None:
            # check velocity constraint
            violated, _ = max_vel_violation(
                relative_position,
                relative_velocity,
                self.config.velocity_constraint.velocity_threshold,
                self.config.velocity_constraint.threshold_distance,
                self.config.velocity_constraint.mean_motion,
                self.config.velocity_constraint.lower_bound,
                slope=self.config.velocity_constraint.slope,
            )

            done = done and violated

        if done:
            next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE

        return done


class TerminalRewardSaturationDoneFunctionValidator(DoneFuncBaseValidator):
    """
    A configuration validator for the TerminalRewardSaturationDoneFunction.

    Attributes
    ----------
    limit : float
        cumulative reward value limit. Done triggers when this limit value is reached
    bound: string
        One of 'upper' or 'lower'. Defines whether the limit value should be an upper bound or a lower bound
    reward_functor: string
        Python path to reward function class. Is resolved into a PyObject
    reward_config: dict
        configuration args for the reward function
    """
    limit: float
    bound: typing.Literal["upper", "lower"]
    reward_functor: PyObject
    reward_config: typing.Dict


class TerminalRewardSaturationDoneFunction(DoneFuncBase):
    """
    A done function which checks if a cumulative reward limit
    has been reached.

    Attributes
    ----------
    config: TerminalRewardSaturationDoneFunctionValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.config: TerminalRewardSaturationDoneFunctionValidator

        self.reward_total = 0
        self.reward_function = self.config.reward_functor(
            **self.config.reward_config, agent_name=self.config.agent_name, platform_names=[self.config.platform_name]
        )

    def reset(self):
        self.reward_total = 0
        self.reward_function.reset()

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        TerminalRewardSaturationDoneFunctionValidator
            Config validator for the TerminalRewardSaturationDoneFunction.
        """

        return TerminalRewardSaturationDoneFunctionValidator

    def __call__(
        self,
        observation,
        action,
        next_observation,
        next_state,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        reward_val = self.reward_function(
            observation, action, next_observation, next_state, next_state, observation_space, observation_units
        )

        self.reward_total += reward_val

        done = False

        if self.config.bound == 'upper' and self.reward_total >= self.config.limit:
            limit_reached = True
        elif self.config.bound == 'lower' and self.reward_total <= self.config.limit:
            limit_reached = True
        else:
            limit_reached = False

        done = limit_reached
        if done:
            next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE

        return done



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements the base initializer class.
"""

import abc
import typing

from corl.libraries.units import Quantity
from pydantic import BaseModel


def strip_units(value) -> typing.Any:
    """
    Strips units from an input value
    Note that no conversions take place.

    Supports CoRL and Pint units

    Parameters
    ----------
    value : Any
        value to be stripped on units

    Returns
    -------
    typing.Any
        same as input but with all units stripped
    """

    if isinstance(value, Quantity):
        return value.m
    return value


def strip_units_from_dict(input_dict: typing.Dict) -> typing.Dict:
    """
    Strips units from input dict and returns dict of values only.
    Note that no conversions take place.

    Supports CoRL and Pint units

    Parameters
    ----------
    input_dict : typing.Dict
        dict of parameters to be stripped of units

    Returns
    -------
    dict
        same as input dict but with all units stripped
    """
    out = {}
    for k, v in input_dict.items():
        out[k] = strip_units(v)
    return out


class InitializerValidator(BaseModel):
    """
    Base Validator for Initializers

    Currently has no parameters, however, make sure to inherit from it to receive future updates to Initializer interface.
    """


class BaseInitializer(abc.ABC):
    """
    This class defines the template for Initializer classes. Initializers are responsible
    for providing a dictionary complete with all relevant agent_reset_config values. Initializers
    encapsulate the initialization of randomized and conditional (dependent) agent state values.
    """

    def __init__(self, config):
        self.config = self.get_validator()(**config)

    @staticmethod
    def get_validator() -> typing.Type[InitializerValidator]:
        """
        get validator for this Done Functor

        Returns:
            DoneFuncBaseValidator -- validator the done functor will use to generate a configuration
        """
        return InitializerValidator

    def __call__(self, **kwargs) -> typing.Dict:
        return self.compute(**kwargs)

    @abc.abstractmethod
    def compute(self, **kwargs) -> typing.Dict:
        """Computes initialization params from an arbitrary set of input kwargs

        Returns
        -------
        typing.Dict
            initialization params generated from input kwargs
        """
        raise NotImplementedError


class BaseInitializerWithPint(BaseInitializer):
    """Generic initializer that utilize a pint UnitRegistry for quantities and unit conversion
    """


class PassThroughInitializer(BaseInitializer):
    """Initializer to simply pass through input kwargs as initialization parameters
    """

    def compute(self, **kwargs):
        return kwargs


class StripUnitsInitializer(BaseInitializer):
    """Initializer to simply strip units from input kwargs and pass magnitude. Note, does not perform unit checking or conversion"""

    def compute(self, **kwargs):
        return strip_units_from_dict(kwargs)


class Accessor(abc.ABC):
    """Base interface for accessor objects to accesss arbitrary sim values for entity initilizer parameters"""

    def __init__(self):
        ...

    @property
    @abc.abstractmethod
    def dependencies(self) -> typing.Set:
        """
        Returns the set of dependencies of this accessor.
        Allows initialization process to initialize dependencies before utilizing this accessor.

        Returns
        -------
        typing.Set
            set of dependency names for this accessor (e.g. names of other entities)
        """

    @abc.abstractmethod
    def access(self, sim, sim_entities: typing.Dict[str, typing.Any]):
        """Access values in the sim and sim entities

        Parameters
        ----------
        sim
            reference to partially initialized sim from which values may be accessed
        sim_entities : typing.Dict[str, typing.Any]
            dictionary of already initialized entites from which values may be accessed
        """


class AttributeAccessor(Accessor):
    """Base accessor implementation for accessing arbitrary attributes from objects

    Parameters
    ----------
    attribute_name: str
        name of attribute to be queried and returned by accessor. May contain nested attributes chained with "."
            e.g. "first.second.third" accesses obj.first.second.third
    """

    def __init__(self, attribute_name):
        super().__init__()
        self.attribute_name = attribute_name

    def get_attribute(self, obj):
        """gets attribute described in constructor from arbitrary input object

        Parameters
        ----------
        obj : Any
            Object from which attribute is accessed

        Returns
        -------
        Any
            accessed attribute
        """
        output = None
        next_obj = obj
        attribute_name_sections = self.attribute_name.split(".")

        if not attribute_name_sections:
            raise ValueError("Attribute name is empty")

        for attribure_name_section in attribute_name_sections:
            output = getattr(next_obj, attribure_name_section)
            next_obj = output

        return output


class SimAttributeAccessor(AttributeAccessor):
    """Accessor for attributes of parent sim object

    Parameters
    ----------
    attribute_name: str
        name of attribute to be queried and returned by accessor. May contain nested attributes chained with "."
            e.g. "first.second.third" accesses obj.first.second.third
    """

    @property
    def dependencies(self) -> typing.Set:
        return set()

    def access(self, sim, _):
        return self.get_attribute(sim)


class EntityAttributeAccessor(AttributeAccessor):
    """Accessor for accessing the attribute of another entity

    Parameters
    ----------
    attribute_name: str
        name of attribute to be queried and returned by accessor. May contain nested attributes chained with "."
            e.g. "first.second.third" accesses obj.first.second.third
    entity_name : str
        name of entity to retrieve attribute from
    """

    def __init__(self, attribute_name, entity_name):
        super().__init__(attribute_name)
        self.entity_name = entity_name

    @property
    def dependencies(self) -> typing.Set:
        return set((self.entity_name, ))

    def access(self, _, sim_entities: typing.Dict[str, typing.Any]):
        return self.get_attribute(sim_entities[self.entity_name])



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements a gaussian decay reward function.
"""

import logging
import typing

import numpy as np
from corl.libraries.utils import get_wrap_diff
from corl.rewards.base_measurement_operation import BaseMeasurementOperation, BaseMeasurementOperationValidator


class GaussianDecayFromTargetValueValidator(BaseMeasurementOperationValidator):
    """
    A configuration validator for GaussianDecayFromTargetValue

    Attributes
    ----------
    reward scale: scale of this reward, this would be the maximum reward value
                  for a given time step
    eps: the length of the reward curve for the exponential decay, would
         recommend playing with this value in a plotting software to determine
         the value you need
    target_value: the value with which to take the difference from the wrapped
                  observation value
    index: the index with which to pull data out of the observation extractor,
           useful if len(observation) > 1
    is_wrap: if the obs difference needs to wrap around 0/360
    is_rad: if the obs difference is in terms of radians
    closer: to only reward with this reward if the difference to the target
            value is less than the last time step
    closer_tolerance: the difference tolerance at which point the agent is close
                      enough to the target value that "closer" is not a concern
    max_diff: reward is zero if absolute difference from target is greater than
              this quantity. Infinite by default.
    """
    reward_scale: float
    eps: float
    target_value: typing.Optional[float] = 0
    index: typing.Optional[int] = 0
    is_wrap: typing.Optional[bool] = False
    is_rad: typing.Optional[bool] = False
    method: typing.Optional[bool] = False
    closer: typing.Optional[bool] = False
    closer_tolerance: typing.Optional[float] = 0.0
    max_diff: float = np.inf


class GaussianDecayFromTargetValue(BaseMeasurementOperation):
    """
    Gaussian Decay from Target Value wraps some sort of observation and takes
    in a target value, the reward based on the difference between the target
    value and the observation. the reward Gaussian decays in value the further
    you are away from the target value.
    """

    @staticmethod
    def get_validator() -> typing.Type[GaussianDecayFromTargetValueValidator]:
        return GaussianDecayFromTargetValueValidator

    def __init__(self, **kwargs) -> None:
        self.config: GaussianDecayFromTargetValueValidator
        super().__init__(**kwargs)
        self._last_value = None
        self._logger = logging.getLogger(self.name)
        self._reward = 0.0

    def __call__(self, observation, action, next_observation, state, next_state, observation_space, observation_units) -> float:
        self._reward = 0.0

        if self.config.agent_name not in next_observation:
            return self._reward

        obs = self.extractor.value(next_observation[self.config.agent_name])[self.config.index]

        if self.config.is_wrap:
            # Note: this always returns the min angle diff and is positive
            diff = get_wrap_diff(obs, self.config.target_value, self.config.is_rad, self.config.method)
        else:
            diff = obs.m - self.config.target_value

        abs_diff = abs(diff)
        if self._last_value is None:
            self._last_value = abs_diff  # type: ignore

        func_applied = 0
        if not self.config.closer or ((self._last_value >= abs_diff) or abs_diff < self.config.closer_tolerance):  # type: ignore
            if not abs_diff > self.config.max_diff:
                func_applied = np.exp(-np.abs(diff**2 / self.config.eps))

        self._reward = self.config.reward_scale * func_applied

        self._last_value = abs_diff  # type: ignore
        return self._reward



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning Core (CoRL) Safe Autonomy Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements the Reward Functions and Reward Validators specific to the inspection task.
"""
import typing
from collections import OrderedDict

import numpy as np
from corl.libraries.state_dict import StateDict
from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator
from corl.simulators.common_platform_utils import get_platform_by_name

from safe_autonomy_sims.utils import get_closest_fft_distance, get_relative_position


class ObservedPointsRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for ObservedPointsReward.

    Attributes
    ----------
    scale : float
        A scalar value applied to reward.
    inspection_entity_name: str
        The name of the entity under inspection.
    weighted_priority : bool
        True to base reward off of point weights, False to base reward off of number of points
    """
    scale: float
    inspection_entity_name: str = "chief"
    weighted_priority: bool = False


class ObservedPointsReward(RewardFuncBase):
    """
    Calculates reward based on the number of new points inspected by the agent.
    """

    def __init__(self, **kwargs):
        self.config: ObservedPointsRewardValidator
        super().__init__(**kwargs)
        self.previous_num_points_inspected = 0
        if self.config.weighted_priority:
            self.previous_weight_inspected = 0.

    @staticmethod
    def get_validator():
        return ObservedPointsRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = 0.0

        inspection_points = next_state.inspection_points_map[self.config.inspection_entity_name]
        current_num_points_inspected = inspection_points.get_num_points_inspected()
        num_new_points = current_num_points_inspected - self.previous_num_points_inspected
        self.previous_num_points_inspected = current_num_points_inspected

        if self.config.weighted_priority:
            current_weight_inspected = inspection_points.get_total_weight_inspected()
            new_weight = current_weight_inspected - self.previous_weight_inspected
            self.previous_weight_inspected = current_weight_inspected
            reward = self.config.scale * new_weight
        else:
            reward = self.config.scale * num_new_points

        return reward


class ChiefDistanceRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for ChiefDistanceReward.

    Attributes
    ----------
    scale : float
        A scalar value applied to reward.
    punishment_reward : float
        Negative scalar associated with the distance done condition.
    max_dist : float
        Maximum allowable distance to chief to recieve reward.
    threshold_dist : float
        Threshold distance
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    """

    scale: float
    punishment_reward: float
    threshold_dist: float
    max_dist: float
    reference_position_sensor_name: str = "reference_position"


class ChiefDistanceReward(RewardFuncBase):
    """
    Calculates reward based on the distance from chief.
    """

    def __init__(self, **kwargs):
        self.config: ChiefDistanceRewardValidator
        super().__init__(**kwargs)
        self.dist_prev = 0.

    @staticmethod
    def get_validator():
        return ChiefDistanceRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = 0.0

        platform = get_platform_by_name(next_state, self.config.platform_names[0])
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        dist = np.linalg.norm(relative_position)

        # Soft constraint
        if dist >= self.config.threshold_dist:
            reward = -self.config.scale * (np.sign(dist - self.dist_prev))

        if dist >= self.config.max_dist:
            reward = self.config.punishment_reward

        self.dist_prev = dist
        return reward


class InspectionDeltaVRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for DockingDeltaVReward

    Attributes
    ----------
    bias : float
        A bias value added to the reward.
    step_size : float
        Size of a single simulation step.
    mass : float
        The mass (kg) of the agent's spacecraft.
    mode : str
        Type of delta-v penalty, either "scale" or "linear_increasing"
    rate : float
        rate at which penalty increases for linear_increasing
    constant_scale : float
        scalar penalty multiplied by delta-v.
        If None, increasing scale value is taken from simulator
    """
    bias: float = 0.0
    step_size: float
    mass: float
    mode: str = 'scale'
    rate: float = 0.0005
    constant_scale: typing.Union[None, float] = None


class InspectionDeltaVReward(RewardFuncBase):
    """
    Calculates reward based on the agent's fuel consumption measured in delta-v.
    """

    def __init__(self, **kwargs):
        self.config: InspectionDeltaVRewardValidator
        super().__init__(**kwargs)
        self.bias = self.config.bias
        self.mass = self.config.mass
        self.step_size = self.config.step_size
        self.mode = self.config.mode
        self.rate = self.config.rate
        self.constant_scale = self.config.constant_scale
        self.scale = 0.0

    def delta_v(self, state):
        """
        Get change in agent's velocity from the current state.

        Parameters
        ----------
        state: StateDict
            The current state of the system.

        Returns
        -------
        d_v: float
            The agent's change in velocity
        """
        deputy = get_platform_by_name(state, self.config.platform_names[0])  # TODO: assuming 1:1 agent:platform
        control_vec = deputy.get_applied_action().m
        d_v = np.sum(np.abs(control_vec)) / self.mass * self.step_size
        return d_v

    def linear_scalar(self, time):
        """
        Delta-v penalty increases linearly with training iteration
        """
        inc_scalar = time * self.rate * self.scale
        return inc_scalar

    @staticmethod
    def get_validator():
        return InspectionDeltaVRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        if self.constant_scale is None:
            self.scale = state.delta_v_scale
        else:
            self.scale = self.constant_scale
        reward = 0.0
        if self.mode == "scale":
            reward = self.scale * self.delta_v(next_state) + self.bias
        elif self.mode == "linear_increasing":
            reward = self.linear_scalar(state.sim_time) * self.delta_v(next_state) + self.bias
        else:
            raise ValueError('mode must be either "scale" or "linear_increasing"')

        return reward


class InspectionSuccessRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the InspectionSuccessReward Reward Function.

    Attributes
    ----------
    scale : float
        Scalar value to adjust magnitude of the reward.
    inspection_entity_name: str
        The name of the entity under inspection.
    weight_threshold : float
        Points score value indicating success.
        By default None, so success occurs when all points are inspected
    """
    scale: float
    inspection_entity_name: str = "chief"
    weight_threshold: typing.Union[float, None] = None


class InspectionSuccessReward(RewardFuncBase):
    """
    This Reward Function is responsible for calculating the reward associated with a successful inspection.
    """

    def __init__(self, **kwargs) -> None:
        self.config: InspectionSuccessRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return InspectionSuccessRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = 0.0

        if self.config.weight_threshold is not None:
            weight = next_state.inspection_points_map[self.config.inspection_entity_name].get_total_weight_inspected()
            if weight >= self.config.weight_threshold:
                reward = self.config.scale

        else:
            inspection_points = next_state.inspection_points_map[self.config.inspection_entity_name]
            all_inspected = all(inspection_points.points_inspected_dict.values())

            if all_inspected:
                reward = self.config.scale

        return reward


class SafeInspectionSuccessRewardValidator(InspectionSuccessRewardValidator):
    """
    A configuration validator for SafeInspectionSuccessReward

    Attributes
    ----------
    mean_motion : float
        orbital mean motion in rad/s of current Hill's reference frame
    crash_region_radius : float
        The radius of the crashing region in meters.
    fft_time_step : float
        Time step to compute the FFT trajectory. FFT is computed for 1 orbit.
    crash_scale : float
        Scalar reward value in the event of a future crash
    """
    mean_motion: float
    crash_region_radius: float
    fft_time_step: float = 1
    crash_scale: float


class SafeInspectionSuccessReward(InspectionSuccessReward):
    """
    This Reward Function is responsible for calculating the reward associated with a successful inspection.
    Considers if a Free Flight Trajectory once the episode ends would result in a collision.
    """

    def __init__(self, **kwargs) -> None:
        self.config: SafeInspectionSuccessRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return SafeInspectionSuccessRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = super().__call__(observation, action, next_observation, state, next_state, observation_space, observation_units)

        if reward != 0.0:
            name = [p for p in self.config.platform_names if p in self.config.agent_name][0]
            pos = next_state.sim_platforms[name].position
            vel = next_state.sim_platforms[name].velocity
            state = np.concatenate((pos, vel))
            n = self.config.mean_motion
            times = np.arange(0, 2 * np.pi / n, self.config.fft_time_step)
            dist = get_closest_fft_distance(state, self.config.mean_motion, times)
            if dist < self.config.crash_region_radius:
                reward = self.config.crash_scale

        return reward


class InspectionCrashRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the InspectionCollisionRewardValidator Reward Function.

    Attributes
    ----------
    scale : float
        Scalar value to adjust magnitude of the reward.
    crash_region_radius : float
        The radius of the crashing region in meters.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    """
    scale: float
    crash_region_radius: float
    reference_position_sensor_name: str = "reference_position"


class InspectionCrashReward(RewardFuncBase):
    """
    This Reward Function is responsible for calculating the reward (penalty) associated with a collision.
    """

    def __init__(self, **kwargs) -> None:
        self.config: InspectionCrashRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return InspectionCrashRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = 0.0

        # Get relatative position + velocity between platform and docking region
        platform = get_platform_by_name(next_state, self.config.platform_names[0])
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        distance = np.linalg.norm(relative_position)

        in_crash_region = distance <= self.config.crash_region_radius

        if in_crash_region:
            reward = self.config.scale

        return reward


class MaxDistanceRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the InspectionCollisionRewardValidator Reward Function.

    Attributes
    ----------
    scale : float
        Scalar value to adjust magnitude of the reward.
    crash_region_radius : float
        The radius of the crashing region in meters.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    """
    scale: float = 1.0
    max_distance: float
    reference_position_sensor_name: str = "reference_position"


class MaxDistanceReward(RewardFuncBase):
    """
    This Reward Function is responsible for calculating the reward (penalty) associated with a collision.
    """

    def __init__(self, **kwargs) -> None:
        self.config: InspectionCrashRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return MaxDistanceRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = 0.0

        # Get relatative position + velocity between platform and docking region
        platform = get_platform_by_name(next_state, self.config.platform_names[0])
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        distance = np.linalg.norm(relative_position)

        out_of_bounds = distance > self.config.max_distance

        if out_of_bounds:
            reward = -self.config.scale

        return reward



"""Reward functions for the inspection tasks"""

import numpy as np
import safe_autonomy_simulation.sims.inspection as sim
from scipy.spatial.transform import Rotation

from safe_autonomy_sims.pettingzoo.inspection.utils import delta_v, rel_dist


def observed_points_reward(chief: sim.Target, prev_num_inspected: int) -> float:
    """A dense reward which rewards the agent for inspecting
    new points during each step of the episode.

    $r_t = 0.01 * (p_t - p_{t-1})$

    where $p_t$ is the total number of inspected points at
    time $t$.

    Parameters
    ----------
    chief : Target
        chief spacecraft under inspection
    prev_num_inspected : int
        number of previously inspected points

    Returns
    -------
    float
        reward value
    """
    current_num_inspected = chief.inspection_points.get_num_points_inspected()

    assert current_num_inspected >= prev_num_inspected, "previously inspected points cannot be uninspected"

    step_inspected = current_num_inspected - prev_num_inspected
    r = 0.1 * step_inspected
    return r


def weighted_observed_points_reward(chief: sim.Target, prev_weight_inspected: float) -> float:
    """A dense reward which rewards the agent for inspecting
    new points during each step of the episode conditioned by
    individual point weights.

    $r_t = 1.0 * (w_t - w_{t-1})$

    where $w_t$ is the total weight of inspected points at
    time $t$.

    Parameters
    ----------
    chief : Target
        chief spacecraft under inspected
    prev_weight_inspected : float
        weight of previously inspected points

    Returns
    -------
    float
        reward value
    """
    current_weight_inspected = chief.inspection_points.get_total_weight_inspected()

    assert current_weight_inspected >= prev_weight_inspected, "previously inspected points cannot be uninspected"

    step_inspected = current_weight_inspected - prev_weight_inspected

    r = 1.0 * step_inspected
    return r


def inspection_success_reward(chief: sim.Target, total_points: int) -> float:
    """A sparse reward applied when the agent successfully
    inspects every point.

    $r_t = 1 if p_t == p_{total}, else 0$

    where $p_t$ is the number of inspected points at time
    $t$ and $p_{total}$ is the total number of points to be
    inspected.

    Parameters
    ----------
    chief : Target
        chief spacecraft under inspection
    total_points : int
        total number of points to be inspected

    Returns
    -------
    float
        reward value
    """
    num_inspected = chief.inspection_points.get_num_points_inspected()
    if num_inspected == total_points:
        r = 1.0
    else:
        r = 0.0
    return r


# pylint:disable=W1401
def weighted_inspection_success_reward(chief: sim.Target, total_weight: float):
    r"""A sparse reward applied when the agent successfully inspects
    point weights above the given threshold.

    $r_t = 1 if w_t \geq w_s else 0$

    where $w_t$ is the total weight of inspected points at time $t$
    and $w_s$ is the total weight of inspected points required for
    successful inspection.

    Parameters
    ----------
    chief : Target
        spacecraft under inspection
    total_weight : float
        inspected weight threshold for success

    Returns
    -------
    float
        reward value
    """
    weight_inspected = chief.inspection_points.get_total_weight_inspected()
    if weight_inspected >= total_weight:
        r = 1.0
    else:
        r = 0.0
    return r


# pylint:disable=W1401
def delta_v_reward(v: np.ndarray, prev_v: np.ndarray, m: float = 12.0, b: float = 0.0):
    r"""A dense reward based on the deputy's fuel
    use (change in velocity).

    $r_t = -((\deltav / m) + b)$

    where
    * $\deltav$ is the change in velocity
    * $m$ is the mass of the deputy
    * $b$ is a tunable bias term

    Parameters
    ----------
    v : np.ndarray
        current velocity
    prev_v : np.ndarray
        previous velocity
    m : float, optional
        deputy mass, by default 12.0
    b : float, optional
        bias term, by default 0.0

    Returns
    -------
    float
        reward value
    """
    r = -((abs(delta_v(v=v, prev_v=prev_v)) / m) + b)
    return r


def crash_reward(chief: sim.Target, deputy: sim.Inspector, crash_radius: float):
    """A sparse reward that punishes the agent
    for intersecting with the chief (crashing).

    $r_t = d_c < r$

    where $d_c$ is the distance between the deputy and
    the chief and $r$ is the radius of the crash region.

    Parameters
    ----------
    chief : Target
        chief spacecraft under inspection
    deputy : Inspector
        deputy spacecraft performing inspection
    crash_radius : float
        distance from chief which triggers a crash

    Returns
    -------
    float
        reward value
    """
    if rel_dist(pos1=chief.position, pos2=deputy.position) < crash_radius:
        r = -1.0
    else:
        r = 0
    return r


# pylint:disable=W1401
def facing_chief_reward(chief: sim.Target, deputy: sim.Inspector, epsilon: float):
    r"""A dense gaussian decaying reward which reward the agent
    for facing the chief.

    $r_t = 0.0005 * e^(-|\delta(f, 1)^2 / \espilon|)$

    where
    * $\delta(f, 1)$ is the difference between 1 and the dot
    product of the camera orientation and the relative position
    between the deputy and the chief
    * $\epsilon$ is the length of the exponential decay curve

    Parameters
    ----------
    chief : Target
        chief spacecraft under inspection
    deputy : Inspector
        deputy spacecraft performing inspection
    epsilon : float
        length of gaussian decay curve

    Returns
    -------
    float
        reward value
    """
    rel_pos = chief.position - deputy.position
    rel_pos = rel_pos / np.linalg.norm(rel_pos)
    gaussian_decay = np.exp(-np.abs(((np.dot(
        Rotation.from_quat(deputy.camera.orientation).as_euler("XYZ"),
        rel_pos,
    ) - 1)**2) / epsilon))
    reward = 0.0005 * gaussian_decay
    return reward


def live_timestep_reward(t: int, t_max: int):
    """A dense reward which rewards the agent for
    each timestep it remains active in the simulation.

    $r_t = 0.001 if t < t_{max}$

    where $t$ is the current time step and $t_{max}$
    is the maximum allowable time for the episode.

    Parameters
    ----------
    t : int
        current time step
    t_max : int
        max time step allowed for episode

    Returns
    -------
    float
        reward value
    """
    reward = 0.0
    if t < t_max:
        reward = 0.001
    return reward



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements evaluation metrics for the inspection environment.
"""
import typing

import numpy as np
from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.metrics.generator import MetricGeneratorTerminalEventScope
from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.types.terminals.real import Real
from corl.evaluation.metrics.types.terminals.void import Void


class DeltaV(MetricGeneratorTerminalEventScope):
    """Generates vector of Dicts for the control for each step during an event for an agent
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        arr = 0
        for step in params.steps:

            if agent_id not in step.agents or step.agents[agent_id] is None:
                break

            map_act = step.agents[agent_id].actions

            if map_act is None:
                continue
            # Create a non terminal metric (Dict) that is comprised of the terminal (Real) actions
            real_dict: typing.Dict[str, Metric] = {key: Real(map_act[key]) for key in map_act.keys()}
            arr += np.sum([abs(c.value[0]) for k, c in real_dict.items() if 'thrust' in k]) / 12 * 10  # TODO: HARDCODED

        return Real(arr)


class InspectedPoints(MetricGeneratorTerminalEventScope):
    """Generates single Real indicating the total inspected points for an event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        # Find the last step with observation data from the platform in question
        # If the platform isn't in the episode at all, return Void
        steps_with_platform_in_question = [item for item in params.steps if agent_id in item.agents and item.agents[agent_id] is not None]
        if len(steps_with_platform_in_question) == 0:
            return Void()

        last_step_with_platform_data = steps_with_platform_in_question[-1]

        if last_step_with_platform_data.agents[agent_id] is None:
            raise RuntimeError("Non Op")

        return Real(last_step_with_platform_data.platforms[0]['sensors']['Sensor_InspectedPoints']['measurement'][0])


class EpisodeLength(MetricGeneratorTerminalEventScope):
    """Generates single Real indicating the episode length for an event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        # Find the last step with observation data from the platform in question
        # If the platform isn't in the episode at all, return Void
        steps_with_platform_in_question = [item for item in params.steps if agent_id in item.agents and item.agents[agent_id] is not None]
        if len(steps_with_platform_in_question) == 0:
            return Void()

        return Real(len(steps_with_platform_in_question))


class Success(MetricGeneratorTerminalEventScope):
    """Generates single Real indicating the success percentage for an event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        # Find the last step with observation data from the platform in question
        # If the platform isn't in the episode at all, return Void
        steps_with_platform_in_question = [item for item in params.steps if agent_id in item.agents and item.agents[agent_id] is not None]
        if len(steps_with_platform_in_question) == 0:
            return Void()

        if params.dones['blue0']['SuccessfulInspectionDoneFunction']:  # TODO: Hardcoded platform name
            val = 1
        else:
            val = 0

        return Real(val)


class SafeSuccess(MetricGeneratorTerminalEventScope):
    """Generates single Real indicating the success percentage for an event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        # Find the last step with observation data from the platform in question
        # If the platform isn't in the episode at all, return Void
        steps_with_platform_in_question = [item for item in params.steps if agent_id in item.agents and item.agents[agent_id] is not None]
        if len(steps_with_platform_in_question) == 0:
            return Void()

        if params.dones['blue0']['SafeSuccessfulInspectionDoneFunction']:  # TODO: Hardcoded platform name
            val = 1
        else:
            val = 0

        return Real(val)


class InspectedPointsScore(MetricGeneratorTerminalEventScope):
    """Generates single Real indicating the score for inspected points score for an event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        # Find the last step with observation data from the platform in question
        # If the platform isn't in the episode at all, return Void
        steps_with_platform_in_question = [item for item in params.steps if agent_id in item.agents and item.agents[agent_id] is not None]
        if len(steps_with_platform_in_question) == 0:
            return Void()

        last_step_with_platform_data = steps_with_platform_in_question[-1]

        if last_step_with_platform_data.agents[agent_id] is None:
            raise RuntimeError("Non Op")

        return Real(last_step_with_platform_data.platforms[0]['sensors']['Sensor_InspectedPointsScore']['measurement'][0])



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements CoRL experiments for safe autonomy training
runs.

CoRL `Experiment`s act as containers around training runs. Within
an experiment a user can define custom training behavior and
develop integration for different RL frameworks and backend simulator
solutions.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements agents capable of using run-time-assurance.
"""

import typing

from corl.agents.base_agent import BaseAgentParser, Functor, FunctorWrapper, TrainableBaseAgent


class FunctorMultiWrapper(Functor):
    """
    A wrapper around one or more functors.

    Attributes
    ----------
    wrapped: Functor, FunctorWrapper
        The functor or functor wrapper configuration wrapped by this functor wrapper.
    """
    wrapped_glues: typing.Union['FunctorMultiWrapper', FunctorWrapper, Functor]

    def create_functor_glue_object(self, platform, agent_name, local_param_storage, world_param_storage):
        """
        Creates a glue functor wrapper object associated with a platform and an agent.

        Parameters
        ----------
        platform : BasePlatform
            The platform instance to be associated with the glue.
        agent_name : str
            The name of the agent to be associated with the glue.
        local_param_storage : [type]
            [description]
        world_param_storage : [type]
            [description]

        Returns
        -------
        MultiWrapperGlue:
            An instance of a glue functor wrapper initialized with the given config.
        """
        wrapped_glues = [
            glue.create_functor_glue_object(platform, agent_name, local_param_storage, world_param_storage) for glue in self.wrapped_glues
        ]
        kwargs = self.config
        kwargs["name"] = self.name
        kwargs["wrapped_glues"] = wrapped_glues
        kwargs["platform"] = platform
        kwargs["agent_name"] = agent_name
        return self.functor(**kwargs)


FunctorMultiWrapper.update_forward_refs()


class ExtendedAgentParser(BaseAgentParser):
    """
    Agent parser extension which parses glue objects.

    Attributes
    ----------
    glues: list
        List of wrapped glues
    """
    glues: typing.List[typing.Union[FunctorMultiWrapper, FunctorWrapper, Functor]]


class RTAAgent(TrainableBaseAgent):
    """
    A trainable RTA Agent

    Attributes
    ----------
    config: ExtendedAgentParser
        The agent configuration
    """

    def __init__(self, **kwargs) -> None:
        self.config: ExtendedAgentParser
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return ExtendedAgentParser
